/*
 *   SONEWS News Server
 *   see AUTHORS for the list of contributors
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.sonews.storage;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.mail.Header;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.Session;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import org.sonews.util.io.Resource;

/**
 * This is MimeMessage which enables custom Message-ID header
 * (this header will not be overwritten by the default one like in MimeMessage).
 * 
 * Also add header and body separate serialization.
 * 
 * And can be deserialized from SQL ResultSet
 * 
 * @author František Kučera (frantovo.cz)
 */
public class DrupalMessage extends MimeMessage {

	private static final Logger log = Logger.getLogger(DrupalMessage.class.getName());
	private static final String MESSAGE_ID_HEADER = "Message-ID";
	private static final String CRLF = "\r\n";
	public static final String CHARSET = "UTF-8";
	private static final String XHTML_CONTENT_TYPE = "text/html; charset=" + CHARSET;
	private String messageID;

	/**
	 * Constructs MIME message from SQL result.
	 * @param rs ResultSet containing message data. No {@link ResultSet#next()} will be called, just values from current row will be read.
	 * @param constructBody true if whole message should be constructed | false if we need only message headers (body will be dummy).
	 */
	public DrupalMessage(ResultSet rs, String myDomain, boolean constructBody) throws SQLException, UnsupportedEncodingException, MessagingException {
		super(Session.getDefaultInstance(System.getProperties()));

		addHeader("Message-id", constructMessageId(rs.getInt("id"), rs.getInt("group_id"), rs.getString("group_name"), myDomain));
		addHeader("Newsgroups", rs.getString("group_name"));
		setFrom(new InternetAddress(rs.getString("sender_email"), rs.getString("sender_name")));
		setSubject(rs.getString("subject"));
		setSentDate(new Date(rs.getLong("created")));

		int parentID = rs.getInt("parent_id");
		if (parentID > 0) {
			String parentMessageID = constructMessageId(parentID, rs.getInt("group_id"), rs.getString("group_name"), myDomain);
			addHeader("In-Reply-To", parentMessageID);
			addHeader("References", parentMessageID);
		}

		if (constructBody) {
			Multipart multipart = new MimeMultipart("alternative");
			setContent(multipart);

			/** XHTML part */
			MimeBodyPart htmlPart = new MimeBodyPart();
			String xhtmlText = readXhtmlText(rs);
			htmlPart.setContent(xhtmlText, XHTML_CONTENT_TYPE);

			/** Plain text part */
			MimeBodyPart textPart = new MimeBodyPart();
			String plainText = readPlainText(rs, xhtmlText);
			textPart.setText(plainText);
			//addHeader("Lines", String.valueOf(plainText.split("\n").length));

			/**
			 * Thunderbirdu záleží, v jakém pořadí části jsou 
			 * (když je prostý text druhý, html se nezobrazí),
			 * KNode zobrazuje HTML správně, i když je na prvním místě.
			 */
			multipart.addBodyPart(textPart);
			multipart.addBodyPart(htmlPart);
		} else {
			/** empty body, just headers */
			setText("");
		}
	}

	private String readPlainText(ResultSet rs, String xhtmlText) {
		try {
			TransformerFactory tf = TransformerFactory.newInstance();
			Transformer textTransformer = tf.newTransformer(new StreamSource(Resource.getAsStream("helpers/mimeTextPart.xsl")));

			StringReader input = new StringReader(xhtmlText);
			StringWriter output = new StringWriter(xhtmlText.length());
			textTransformer.transform(new StreamSource(input), new StreamResult(output));

			return output.toString();
		} catch (Exception e) {
			/**
			 * TODO: lepší ošetření chyby
			 */
			log.log(Level.WARNING, "Error while transforming article to plain text", e);
			return makeSimpleXHTML("Při transformaci příspěvku bohužel došlo k chybě.");
		}
	}

	private String readXhtmlText(ResultSet rs) {
		/**
		 * TODO: 
		 *		- znovupoužívat XSL transformér
		 *		- používat cache, ukládat si vygenerované články
		 */
		try {
			String inputText = makeSimpleXHTML(rs.getString("text"));

			TransformerFactory tf = TransformerFactory.newInstance();
			Transformer paragraphTransformer = tf.newTransformer(new StreamSource(Resource.getAsStream("helpers/mimeXhtmlPart-make-paragraphs.xsl")));

			String paragraphedText;
			boolean tidyWasUsed = false;
			try {
				StringReader input = new StringReader(inputText);
				StringWriter output = new StringWriter(2 * inputText.length());
				paragraphTransformer.transform(new StreamSource(input), new StreamResult(output));
				paragraphedText = output.toString();
			} catch (Exception e) {
				log.log(Level.FINER, "HTML input was shitty – Tidy had to be called.", e);
				StringReader input = new StringReader(tidyXhtml(inputText));
				StringWriter output = new StringWriter(2 * inputText.length());
				paragraphTransformer.transform(new StreamSource(input), new StreamResult(output));
				paragraphedText = output.toString();
				tidyWasUsed = true;
			}

			Transformer xhtmlTransformer = tf.newTransformer(new StreamSource(Resource.getAsStream("helpers/mimeXhtmlPart.xsl")));
			xhtmlTransformer.setParameter("isRoot", (rs.getInt("parent_id") == 0));
			xhtmlTransformer.setParameter("title", rs.getString("subject"));
			xhtmlTransformer.setParameter("urlBase", rs.getString("urlBase"));
			xhtmlTransformer.setParameter("wwwRead", rs.getString("wwwRead"));
			xhtmlTransformer.setParameter("wwwPost", rs.getString("wwwPost"));
			xhtmlTransformer.setParameter("headComment", String.format("Drupal-NNTP bridge. Transformed: %1$tc. Tidy had to be used: %2$b", new Date(), tidyWasUsed));
			StringReader input = new StringReader(paragraphedText);
			StringWriter output = new StringWriter(2 * paragraphedText.length());
			xhtmlTransformer.transform(new StreamSource(input), new StreamResult(output));

			return output.toString();
		} catch (Exception e) {
			/**
			 * TODO: lepší ošetření chyby
			 */
			log.log(Level.WARNING, "Error while transforming article to XHTML", e);
			return makeSimpleXHTML("<p>Při transformaci příspěvku bohužel došlo k chybě.</p>");
		}
	}

	private static String makeSimpleXHTML(String body) {
		return "<html xmlns=\"http://www.w3.org/1999/xhtml\"><body>" + body + "</body></html>";
	}

	/**
	 * TODO: refaktorovat, přesunout
	 */
	private static String tidyXhtml(String inputText) throws IOException {
		/*
		 * Viz https://sourceforge.net/tracker/index.php?func=detail&aid=3424437&group_id=27659&atid=390966
		 *
		 * TODO:
		 *		- použít delší zástupný řetězec, ne jen jeden znak
		 *		- umísťovat ho jen tam, kde už nějaký text je (ne mezi >\s*<)
		 */
		inputText = inputText.replaceAll("\\n", "◆\n");

		Runtime r = Runtime.getRuntime();
		Process p = r.exec(new String[]{"tidy", // http://tidy.sourceforge.net
					"-asxml", // well formed XHTML
					"-numeric", // číselné entity
					"-utf8", // kódování
					"--show-warnings", "false", // žádná varování nás nezajímají
					"--show-errors", "0", // ani chyby
					"--doctype", "omit", // doctype nepotřebujeme (doplníme si případně vlastní v XSLT)
					"--logical-emphasis", "true", // em a strong místo i a b
					"--literal-attributes", "true", // zachovat mezery a konce řádků v atributech
					"--force-output", "true" // neznámé značky zahodíme, vložíme jen jejich obsah
				});

		PrintStream vstupProcesu = new PrintStream(p.getOutputStream());
		vstupProcesu.print(inputText);
		vstupProcesu.close();

		String outputText = streamToString(p.getInputStream());

		outputText = outputText.replaceAll("◆\\n", "\n");
		outputText = outputText.replaceAll("◆", "\n");

		return outputText;
	}

	/**
	 * TODO: refaktorovat, přesunout
	 */
	private static String streamToString(InputStream proud) throws IOException {
		StringBuilder výsledek = new StringBuilder();
		BufferedReader buf = new BufferedReader(new InputStreamReader(proud));
		while (true) {
			String radek = buf.readLine();
			if (radek == null) {
				break;
			} else {
				výsledek.append(radek);
				výsledek.append("\n");
			}
		}
		return výsledek.toString();
	}

	private static String constructMessageId(int articleID, int groupID, String groupName, String domainName) {
		StringBuilder sb = new StringBuilder();
		sb.append("<");
		sb.append(articleID);
		sb.append("-");
		sb.append(groupID);
		sb.append("-");
		sb.append(groupName);
		sb.append("@");
		sb.append(domainName);
		sb.append(">");
		return sb.toString();
	}

	@Override
	public void setHeader(String name, String value) throws MessagingException {
		super.setHeader(name, value);

		if (MESSAGE_ID_HEADER.equalsIgnoreCase(name)) {
			messageID = value;
		}
	}

	@Override
	public final void addHeader(String name, String value) throws MessagingException {
		super.addHeader(name, value);

		if (MESSAGE_ID_HEADER.equalsIgnoreCase(name)) {
			messageID = value;
		}
	}

	@Override
	public void removeHeader(String name) throws MessagingException {
		super.removeHeader(name);

		if (MESSAGE_ID_HEADER.equalsIgnoreCase(name)) {
			messageID = null;
		}
	}

	public void setMessageID(String messageID) {
		this.messageID = messageID;
	}

	@Override
	protected void updateMessageID() throws MessagingException {
		if (messageID == null) {
			super.updateMessageID();
		} else {
			setHeader(MESSAGE_ID_HEADER, messageID);
		}
	}

	/**
	 * Call {@link #saveChanges()} before this method, if you want all headers including such ones like:
	 * 
	 * <pre>MIME-Version: 1.0
	 *Content-Type: multipart/alternative;</pre>
	 * 
	 * @return serialized headers
	 * @throws MessagingException if getAllHeaders() fails
	 */
	public String getHeaders() throws MessagingException {
		StringBuilder sb = new StringBuilder();
		for (Enumeration eh = getAllHeaderLines(); eh.hasMoreElements();) {
			sb.append(eh.nextElement());
			sb.append(CRLF);
		}
		return sb.toString();
	}

	public byte[] getBody() throws IOException, MessagingException {
		saveChanges();

		ArrayList<String> skipHeaders = new ArrayList<String>();
		for (Enumeration eh = getAllHeaders(); eh.hasMoreElements();) {
			Header h = (Header) eh.nextElement();
			skipHeaders.add(h.getName());
		}

		ByteArrayOutputStream baos = new ByteArrayOutputStream(1024);
		writeTo(baos, skipHeaders.toArray(new String[skipHeaders.size()]));
		return baos.toByteArray();
	}
}
